/* The copyright in this software is being made available under the BSD
 * License, included below. This software may be subject to other third party
 * and contributor rights, including patent rights, and no such rights are
 * granted under this license.  
 *
 * Copyright (c) 2010-2011, ITU/ISO/IEC
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the ITU/ISO/IEC nor the names of its contributors may
 *    be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

/** \file     TAppDecTop.cpp
    \brief    Decoder application class
*/

#include <list>
#include <vector>
#include <stdio.h>
#include <fcntl.h>
#include <assert.h>

#include "TAppDecTop.h"
#include "TLibDecoder/AnnexBread.h"
#include "TLibDecoder/NALread.h"

#if DEC_REF
#include <math.h>
#endif
//! \ingroup TAppDecoder
//! \{

// ====================================================================================================================
// Constructor / destructor / initialization / destroy
// ====================================================================================================================

TAppDecTop::TAppDecTop()
{
  ::memset (m_abDecFlag, 0, sizeof (m_abDecFlag));
  m_iPOCLastDisplay  = -1;
}

Void TAppDecTop::create()
{
}

Void TAppDecTop::destroy()
{
}

// ====================================================================================================================
// Public member functions
// ====================================================================================================================

/**
 - create internal class
 - initialize internal class
 - until the end of the bitstream, call decoding function in TDecTop class
 - delete allocated buffers
 - destroy internal class
 .
 */
Void TAppDecTop::decode()
{
  UInt                uiPOC;
  TComList<TComPic*>* pcListPic = NULL;

  ifstream bitstreamFile(m_pchBitstreamFile, ifstream::in | ifstream::binary);
  if (!bitstreamFile)
  {
    fprintf(stderr, "\nfailed to open bitstream file `%s' for reading\n", m_pchBitstreamFile);
    exit(EXIT_FAILURE);
  }

  InputByteStream bytestream(bitstreamFile);

  // create & initialize internal classes
  xCreateDecLib();
  xInitDecLib  ();
  m_iPOCLastDisplay += m_iSkipFrame;      // set the last displayed POC correctly for skip forward.

  // main decoder loop
  bool recon_opened = false; // reconstruction file not yet opened. (must be performed after SPS is seen)

  while (!!bitstreamFile)
  {
    /* location serves to work around a design fault in the decoder, whereby
     * the process of reading a new slice that is the first slice of a new frame
     * requires the TDecTop::decode() method to be called again with the same
     * nal unit. */
    streampos location = bitstreamFile.tellg();
    AnnexBStats stats = AnnexBStats();

    vector<uint8_t> nalUnit;
    byteStreamNALUnit(bytestream, nalUnit, stats);

    // call actual decoding function
    bool bNewPicture = false;
    if (nalUnit.empty())
      /* this can happen if the following occur:
       *  - empty input file
       *  - two back-to-back start_code_prefixes
       *  - start_code_prefix immediately followed by EOF
       */
      fprintf(stderr, "Warning: Attempt to decode an empty NAL unit\n");
    else
    {
      InputNALUnit nalu;
      read(nalu, nalUnit);
      bNewPicture = m_cTDecTop.decode(nalu, m_iSkipFrame, m_iPOCLastDisplay);
      if (bNewPicture)
      {
        bitstreamFile.clear();
        /* location points to the current nalunit payload[1] due to the
         * need for the annexB parser to read three extra bytes.
         * [1] except for the first NAL unit in the file
         *     (but bNewPicture doesn't happen then) */
        bitstreamFile.seekg(location-streamoff(3));
        bytestream.reset();
      }
    }
    if (bNewPicture || !bitstreamFile)
    {
      m_cTDecTop.executeDeblockAndAlf(uiPOC, pcListPic, m_iSkipFrame, m_iPOCLastDisplay);
    }

    if( pcListPic )
    {
      if ( m_pchReconFile && !recon_opened )
      {
        if ( m_outputBitDepth == 0 )
          m_outputBitDepth = g_uiBitDepth + g_uiBitIncrement;

        m_cTVideoIOYuvReconFile.open( m_pchReconFile, true, m_outputBitDepth, g_uiBitDepth + g_uiBitIncrement ); // write mode
        recon_opened = true;
#if DEC_REF
  if ( m_pchRefFile )
	{
		m_cTVideoIOYuvRefFile.open( m_pchRefFile, false, m_outputBitDepth, g_uiBitDepth + g_uiBitIncrement);					// read mode
	}
#endif
      }
      // write reconstuction to file
      xWriteOutput( pcListPic );
    }
  }
  
  // delete buffers
  m_cTDecTop.deletePicBuffer();
  
  // destroy internal classes
  xDestroyDecLib();
}

// ====================================================================================================================
// Protected member functions
// ====================================================================================================================

Void TAppDecTop::xCreateDecLib()
{
  // create decoder class
  m_cTDecTop.create();
}

Void TAppDecTop::xDestroyDecLib()
{
  if ( m_pchReconFile )
  {
    m_cTVideoIOYuvReconFile. close();
  }
#if DEC_REF
  if ( m_pchRefFile )
	{
		m_cTVideoIOYuvRefFile. close();
	}
#endif     
  // destroy decoder class
  m_cTDecTop.destroy();
}

Void TAppDecTop::xInitDecLib()
{
  // initialize decoder class
  m_cTDecTop.init();
  m_cTDecTop.setPictureDigestEnabled(m_pictureDigestEnabled);
}

/** \param pcListPic list of pictures to be written to file
    \todo            DYN_REF_FREE should be revised
 */
Void TAppDecTop::xWriteOutput( TComList<TComPic*>* pcListPic )
{
  TComList<TComPic*>::iterator iterPic   = pcListPic->begin();
#if DEC_REF
  TComPicYuv* pcPicRef = new TComPicYuv();
#endif   
  while (iterPic != pcListPic->end())
  {
    TComPic* pcPic = *(iterPic);
#if DEC_REF
    // allocate destination buffer
    pcPicRef->create( pcPic->getPicYuvRec()->getWidth (), 
      pcPic->getPicYuvRec()->getHeight(), 
      g_uiMaxCUWidth, 
      g_uiMaxCUHeight, 
      g_uiMaxCUDepth );
#endif        
    if ( pcPic->getReconMark() && pcPic->getPOC() == (m_iPOCLastDisplay + 1) )
    {
      // write to file
      if ( m_pchReconFile )
      {
        m_cTVideoIOYuvReconFile.write( pcPic->getPicYuvRec(), pcPic->getSlice(0)->getSPS()->getPad() );
      }
#if DEC_REF
		// read the ref file
		if ( m_pchRefFile )
		{
			m_cTVideoIOYuvRefFile.read( pcPicRef, pcPic->getSlice(0)->getSPS()->getPad() );
		}
		find_psnr(pcPicRef, pcPic->getPicYuvRec(), pcPic->getSlice(0)->getSPS()->getPad(), pcPic->getSlice(0)->getPOC() );
#endif 

      // update POC of display order
      m_iPOCLastDisplay = pcPic->getPOC();
      
      // erase non-referenced picture in the reference picture list after display
      if ( !pcPic->getSlice(0)->isReferenced() && pcPic->getReconMark() == true )
      {
#if !DYN_REF_FREE
        pcPic->setReconMark(false);
        
        // mark it should be extended later
        pcPic->getPicYuvRec()->setBorderExtension( false );
        
#else
        pcPic->destroy();
        pcListPic->erase( iterPic );
        iterPic = pcListPic->begin(); // to the beginning, non-efficient way, have to be revised!
        continue;
#endif
      }
    }
#if DEC_REF
      // destroy temporary buffer
			pcPicRef->destroy();
#endif 

    iterPic++;
  }
}
#if DEC_REF
Void TAppDecTop::find_psnr(TComPicYuv* pcPicRef, TComPicYuv* pcPicDec, Int aiPad[2], Int POC)
{
  Int64 diff_y,diff_u,diff_v;
  Double snr_y, snr_u, snr_v;
  Int   x, y;
  Pel*  pcRef   = pcPicRef->getLumaAddr();
  Pel*  pcRefCb = pcPicRef->getCbAddr();
  Pel*  pcRefCr = pcPicRef->getCrAddr();
  Pel*  pcDec   = pcPicDec->getLumaAddr();
  Pel*  pcDecCb = pcPicDec->getCbAddr();
  Pel*  pcDecCr = pcPicDec->getCrAddr();

  Int   iStride = pcPicRef->getStride();
  Int   iWidth, iHeight;
  // compute actual YUV width & height excluding padding size
  iWidth  = pcPicRef->getWidth () - aiPad[0];
  iHeight = pcPicRef->getHeight() - aiPad[1];

  diff_y = diff_u = diff_v = 0;
  //Y
  for ( y = 0; y < iHeight; y++ )
  {
    for ( x = 0; x < iWidth; x++ ) 
      diff_y += (pcRef[x]-pcDec[x]) * (pcRef[x]-pcDec[x]);

    pcRef += iStride;
    pcDec += iStride;
  }
  // Collecting SNR statistics
  if (diff_y != 0)
    snr_y=(Double)(10*log10(255*255*(double)((double)(iWidth)*(iHeight) / diff_y)));        // luma snr for current frame
  else
    snr_y=0.0;

  iWidth   >>= 1;
  iHeight  >>= 1;
  iStride  >>= 1;

  //U, V
  for ( y = 0; y < iHeight; y++ )
  {
    for ( x = 0; x < iWidth; x++ ) 
    {
      diff_u += (pcRefCb[x]-pcDecCb[x]) * (pcRefCb[x]-pcDecCb[x]);
      diff_v += (pcRefCr[x]-pcDecCr[x]) * (pcRefCr[x]-pcDecCr[x]);
    }

    pcRefCb += iStride;
    pcDecCb += iStride;
    pcRefCr += iStride;
    pcDecCr += iStride;
  }

  // Collecting SNR statistics
  if (diff_u != 0)
    snr_u=(float)(10*log10(255*255*(double)((double)(iWidth)*(iHeight) / (diff_u))));    //  chroma snr for current frame
  else
    snr_u=0.0;
  if (diff_v != 0)
    snr_v=(float)(10*log10(255*255*(double)((double)(iWidth)*(iHeight) / (diff_v))));    //  chroma snr for current frame
  else
    snr_v=0;
  printf("\nPOC    %d: PSNR Y: %3.4f, U: %3.4f, V: %3.4f\n", POC, snr_y, snr_u, snr_v);
}
#endif
//! \}
